# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
import capellambse.model as m


class LiteralBooleanValue(m.GenericElement):
    """A Literal Boolean Value."""

    _xmltag = "ownedLiterals"

    value = m.BoolPOD("value")


class LiteralValue(m.GenericElement):
    is_abstract = m.BoolPOD("abstract")
    """Indicates if property is abstract."""
    value = m.StringPOD("value")
    type = m.AttrProxyAccessor(m.GenericElement, "abstractType")


@m.xtype_handler(None)
class LiteralNumericValue(LiteralValue):
    value = m.StringPOD("value")
    unit = m.AttrProxyAccessor(m.GenericElement, "unit")


@m.xtype_handler(None)
class LiteralStringValue(LiteralValue):
    """A Literal String Value."""


@m.xtype_handler(None)
class ValuePart(m.GenericElement):
    """A Value Part of a Complex Value."""

    _xmltag = "ownedParts"

    referenced_property = m.AttrProxyAccessor(
        m.GenericElement, "referencedProperty"
    )
    value = m.RoleTagAccessor("ownedValue")


@m.xtype_handler(None)
class ComplexValue(m.GenericElement):
    """A Complex Value."""

    _xmltag = "ownedDataValues"

    type = m.AttrProxyAccessor(m.GenericElement, "abstractType")
    value_parts = m.DirectProxyAccessor(ValuePart, aslist=m.ElementList)


@m.attr_equal("name")
@m.xtype_handler(None)
class EnumerationLiteral(m.GenericElement):
    _xmltag = "ownedLiterals"

    value = m.RoleTagAccessor("domainValue")

    owner: m.Accessor


@m.xtype_handler(None)
class EnumerationReference(m.GenericElement):
    type = m.AttrProxyAccessor(m.GenericElement, "abstractType")
    value = m.AttrProxyAccessor(m.GenericElement, "referencedValue")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Logical Architecture layer.

.. diagram:: [CDB] LA ORM
"""

from __future__ import annotations

from capellambse import model as m

from . import capellacommon, capellacore, cs, fa, interaction, sa


@m.xtype_handler(None)
class LogicalFunction(fa.Function):
    """A logical function on the Logical Architecture layer."""

    realized_system_functions = m.TypecastAccessor(
        sa.SystemFunction, "realized_functions"
    )
    owner: m.Accessor[LogicalComponent]


@m.xtype_handler(None)
class LogicalFunctionPkg(m.GenericElement):
    """A logical function package."""

    _xmltag = "ownedFunctionPkg"

    functions = m.RoleTagAccessor(
        "ownedLogicalFunctions", LogicalFunction, aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class LogicalComponent(cs.Component):
    """A logical component on the Logical Architecture layer."""

    _xmltag = "ownedLogicalComponents"

    allocated_functions = m.LinkAccessor[LogicalFunction](
        "ownedFunctionalAllocation",
        fa.ComponentFunctionalAllocation,
        aslist=m.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )
    realized_system_components = m.TypecastAccessor(
        sa.SystemComponent,
        "realized_components",
    )

    components: m.Accessor


@m.xtype_handler(None)
class LogicalComponentPkg(m.GenericElement):
    """A logical component package."""

    _xmltag = "ownedLogicalComponentPkg"

    components = m.DirectProxyAccessor(LogicalComponent, aslist=m.ElementList)
    state_machines = m.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=m.ElementList
    )
    exchanges = m.DirectProxyAccessor(
        fa.ComponentExchange, aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class CapabilityRealization(m.GenericElement):
    """A capability."""

    _xmltag = "ownedCapabilityRealizations"

    owned_chains = m.DirectProxyAccessor(
        fa.FunctionalChain, aslist=m.ElementList
    )
    involved_functions = m.LinkAccessor[LogicalFunction](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        interaction.AbstractFunctionAbstractCapabilityInvolvement,
        aslist=m.ElementList,
        attr="involved",
    )
    involved_chains = m.LinkAccessor[fa.FunctionalChain](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        interaction.FunctionalChainAbstractCapabilityInvolvement,
        aslist=m.ElementList,
        attr="involved",
    )
    involved_components = m.LinkAccessor[LogicalComponent](
        "ownedCapabilityRealizationInvolvements",
        capellacommon.CapabilityRealizationInvolvement,
        aslist=m.MixedElementList,
        attr="involved",
    )
    realized_capabilities = m.LinkAccessor[sa.Capability](
        "ownedAbstractCapabilityRealizations",
        interaction.AbstractCapabilityRealization,
        aslist=m.ElementList,
        attr="targetElement",
    )

    postcondition = m.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = m.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = m.DirectProxyAccessor(
        interaction.Scenario, aslist=m.ElementList
    )
    states = m.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class CapabilityRealizationPkg(m.GenericElement):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = m.DirectProxyAccessor(
        CapabilityRealization, aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class LogicalArchitecture(cs.ComponentArchitecture):
    """Provides access to the LogicalArchitecture layer of the model."""

    root_component = m.AttributeMatcherAccessor(
        LogicalComponent,
        attributes={"is_actor": False},
        rootelem=LogicalComponentPkg,
    )
    root_function = m.DirectProxyAccessor(
        LogicalFunction, rootelem=LogicalFunctionPkg
    )

    function_package = m.DirectProxyAccessor(LogicalFunctionPkg)
    component_package = m.DirectProxyAccessor(LogicalComponentPkg)
    capability_package = m.DirectProxyAccessor(CapabilityRealizationPkg)

    all_functions = m.DeepProxyAccessor(
        LogicalFunction,
        aslist=m.ElementList,
        rootelem=LogicalFunctionPkg,
    )
    all_capabilities = m.DeepProxyAccessor(
        CapabilityRealization, aslist=m.ElementList
    )
    all_components = (  # maybe this should exclude .is_actor
        m.DeepProxyAccessor(LogicalComponent, aslist=m.ElementList)
    )
    all_actors = property(
        lambda self: self._model.search(LogicalComponent).by_is_actor(True)
    )
    all_functional_chains = property(
        lambda self: self._model.search(fa.FunctionalChain, below=self)
    )

    actor_exchanges = m.DirectProxyAccessor(
        fa.ComponentExchange,
        aslist=m.ElementList,
        rootelem=LogicalComponentPkg,
    )
    component_exchanges = m.DeepProxyAccessor(
        fa.ComponentExchange,
        aslist=m.ElementList,
        rootelem=[LogicalComponentPkg, LogicalComponent],
    )

    all_function_exchanges = m.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=m.ElementList,
        rootelem=[LogicalFunctionPkg, LogicalFunction],
    )
    all_component_exchanges = m.DeepProxyAccessor(
        fa.ComponentExchange, aslist=m.ElementList
    )

    diagrams = m.DiagramAccessor(
        "Logical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )


m.set_accessor(
    sa.Capability,
    "realizing_capabilities",
    m.ReferenceSearchingAccessor(
        CapabilityRealization, "realized_capabilities", aslist=m.ElementList
    ),
)
m.set_accessor(
    sa.SystemComponent,
    "realizing_logical_components",
    m.ReferenceSearchingAccessor(
        LogicalComponent, "realized_components", aslist=m.ElementList
    ),
)
m.set_accessor(
    sa.SystemFunction,
    "realizing_logical_functions",
    m.ReferenceSearchingAccessor(
        LogicalFunction, "realized_system_functions", aslist=m.ElementList
    ),
)
m.set_accessor(
    LogicalFunction,
    "owner",
    m.ReferenceSearchingAccessor(LogicalComponent, "allocated_functions"),
)
m.set_accessor(
    LogicalFunction,
    "packages",
    m.DirectProxyAccessor(
        LogicalFunctionPkg,
        aslist=m.ElementList,
    ),
)
m.set_accessor(
    LogicalFunction,
    "involved_in",
    m.ReferenceSearchingAccessor(
        CapabilityRealization, "involved_functions", aslist=m.ElementList
    ),
)
m.set_self_references(
    (LogicalComponent, "components"),
    (LogicalComponentPkg, "packages"),
    (LogicalFunction, "functions"),
    (LogicalFunctionPkg, "packages"),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Some additional glue to tie ReqIF into capellambse."""

__all__ = ["init"]

import capellambse.model as m
from capellambse.metamodel import cs

from . import _capellareq as cr
from . import _requirements as rq


def init() -> None:
    m.set_accessor(
        rq.Folder,
        "folders",
        m.DirectProxyAccessor(rq.Folder, aslist=m.ElementList),
    )
    m.set_accessor(
        m.GenericElement, "requirements", cr.ElementRelationAccessor()
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "requirement_modules",
        m.DirectProxyAccessor(cr.CapellaModule, aslist=m.ElementList),
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "all_requirements",
        m.DeepProxyAccessor(
            rq.Requirement, aslist=m.ElementList, rootelem=cr.CapellaModule
        ),
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "requirement_types_folders",
        m.DirectProxyAccessor(cr.CapellaTypesFolder, aslist=m.ElementList),
    )
    m.set_accessor(
        cr.CapellaModule,
        "requirement_types_folders",
        m.DirectProxyAccessor(cr.CapellaTypesFolder, aslist=m.ElementList),
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "all_requirement_types",
        m.DeepProxyAccessor(
            rq.RequirementType,
            aslist=m.ElementList,
            rootelem=cr.CapellaTypesFolder,
        ),
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "all_module_types",
        m.DeepProxyAccessor(
            rq.ModuleType, aslist=m.ElementList, rootelem=cr.CapellaTypesFolder
        ),
    )
    m.set_accessor(
        cs.ComponentArchitecture,
        "all_relation_types",
        m.DeepProxyAccessor(
            rq.RelationType,
            aslist=m.ElementList,
            rootelem=cr.CapellaTypesFolder,
        ),
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Classes handling Mode/State-Machines and related content."""

import capellambse.model as m

from . import capellacore, modellingcore


class AbstractStateRealization(m.GenericElement): ...


class TransfoLink(m.GenericElement): ...


class CapabilityRealizationInvolvement(m.GenericElement): ...


@m.xtype_handler(None)
class Region(m.GenericElement):
    """A region inside a state machine or state/mode."""

    _xmltag = "ownedRegions"

    states: m.Accessor
    modes: m.Accessor
    transitions: m.Accessor


class AbstractStateMode(m.GenericElement):
    """Common code for states and modes."""

    _xmltag = "ownedStates"

    regions = m.DirectProxyAccessor(Region, aslist=m.ElementList)


@m.xtype_handler(None)
class State(AbstractStateMode):
    """A state."""

    entries = m.AttrProxyAccessor(
        m.GenericElement, "entry", aslist=m.MixedElementList
    )
    do_activity = m.AttrProxyAccessor(
        m.GenericElement, "doActivity", aslist=m.MixedElementList
    )
    exits = m.AttrProxyAccessor(
        m.GenericElement, "exit", aslist=m.MixedElementList
    )

    functions: m.Accessor


@m.xtype_handler(None)
class Mode(AbstractStateMode):
    """A mode."""


@m.xtype_handler(None)
class DeepHistoryPseudoState(AbstractStateMode):
    """A deep history pseudo state."""


@m.xtype_handler(None)
class FinalState(AbstractStateMode):
    """A final state."""


@m.xtype_handler(None)
class ForkPseudoState(AbstractStateMode):
    """A fork pseudo state."""


@m.xtype_handler(None)
class InitialPseudoState(AbstractStateMode):
    """An initial pseudo state."""


@m.xtype_handler(None)
class JoinPseudoState(AbstractStateMode):
    """A join pseudo state."""


@m.xtype_handler(None)
class ShallowHistoryPseudoState(AbstractStateMode):
    """A shallow history pseudo state."""


@m.xtype_handler(None)
class TerminatePseudoState(AbstractStateMode):
    """A terminate pseudo state."""


@m.xtype_handler(None)
class StateMachine(m.GenericElement):
    """A state machine."""

    _xmltag = "ownedStateMachines"

    regions = m.DirectProxyAccessor(Region, aslist=m.ElementList)


@m.xtype_handler(None)
class StateTransition(m.GenericElement):
    r"""A transition between :class:`State`\ s or :class:`Mode`\ s."""

    _xmltag = "ownedTransitions"

    source = m.AttrProxyAccessor(m.GenericElement, "source")
    destination = m.AttrProxyAccessor(m.GenericElement, "target")
    triggers = m.AttrProxyAccessor(
        m.GenericElement, "triggers", aslist=m.MixedElementList
    )
    effects = m.AttrProxyAccessor(
        m.GenericElement, "effect", aslist=m.MixedElementList
    )
    guard = m.AttrProxyAccessor(capellacore.Constraint, "guard")


@m.xtype_handler(None)
class GenericTrace(modellingcore.TraceableElement):
    """A trace between two elements."""

    @property
    def name(self) -> str:  # type: ignore[override]
        """Return the name."""
        direction = ""
        if self.target is not None:
            direction = f" to {self.target.name} ({self.target.uuid})"

        return f"[{type(self).__name__}]{direction}"


m.set_accessor(
    AbstractStateMode,
    "realized_states",
    m.LinkAccessor(
        None,  # FIXME fill in tag
        AbstractStateRealization,
        aslist=m.ElementList,
        attr="targetElement",
    ),
)
for cls in [
    State,
    Mode,
    DeepHistoryPseudoState,
    FinalState,
    ForkPseudoState,
    InitialPseudoState,
    JoinPseudoState,
    ShallowHistoryPseudoState,
    TerminatePseudoState,
]:
    m.set_accessor(
        cls,
        "realizing_states",
        m.ReferenceSearchingAccessor(
            cls, "realized_states", aslist=m.ElementList
        ),
    )

m.set_accessor(
    Region,
    "states",
    m.RoleTagAccessor(AbstractStateMode._xmltag, aslist=m.ElementList),
)
m.set_accessor(
    Region, "modes", m.DirectProxyAccessor(Mode, aslist=m.ElementList)
)
m.set_accessor(
    Region,
    "transitions",
    m.DirectProxyAccessor(StateTransition, aslist=m.ElementList),
)
m.set_accessor(
    m.GenericElement,
    "traces",
    m.DirectProxyAccessor(GenericTrace, aslist=m.ElementList),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implementation of objects and relations for Functional Analysis.

Functional Analysis objects inheritance tree (taxonomy):

.. diagram:: [CDB] FunctionalAnalysis [Taxonomy]

Functional Analysis object-relations map (ontology):

.. diagram:: [CDB] FunctionalAnalysis [Ontology]
"""

from __future__ import annotations

import capellambse.model as m

from . import capellacommon, capellacore, information, interaction, modeltypes


class ComponentExchangeAllocation(m.GenericElement): ...


class ComponentExchangeFunctionalExchangeAllocation(m.GenericElement): ...


class ComponentFunctionalAllocation(m.GenericElement): ...


@m.xtype_handler(None)
class ControlNode(m.GenericElement):
    """A node with a specific control-kind."""

    _xmltag = "ownedSequenceNodes"

    kind = m.EnumPOD("kind", modeltypes.ControlNodeKind, writable=False)


@m.xtype_handler(None)
class FunctionRealization(m.GenericElement):
    """A realization that links to a function."""

    _xmltag = "ownedFunctionRealizations"


class AbstractExchange(m.GenericElement):
    """Common code for Exchanges."""

    source = m.AttrProxyAccessor(m.GenericElement, "source")
    target = m.AttrProxyAccessor(m.GenericElement, "target")


@m.xtype_handler(None)
class AbstractFunction(m.GenericElement):
    """An AbstractFunction."""

    available_in_states = m.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=m.ElementList
    )


@m.xtype_handler(None)
class FunctionPort(m.GenericElement):
    """A function port."""

    owner = m.ParentAccessor(m.GenericElement)
    exchanges: m.Accessor
    state_machines = m.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=m.ElementList
    )


@m.xtype_handler(None)
class FunctionInputPort(FunctionPort):
    """A function input port."""

    _xmltag = "inputs"

    exchange_items = m.AttrProxyAccessor(
        information.ExchangeItem, "incomingExchangeItems", aslist=m.ElementList
    )


@m.xtype_handler(None)
class FunctionOutputPort(FunctionPort):
    """A function output port."""

    _xmltag = "outputs"

    exchange_items = m.AttrProxyAccessor(
        information.ExchangeItem, "outgoingExchangeItems", aslist=m.ElementList
    )


class Function(AbstractFunction):
    """Common Code for Function's."""

    _xmltag = "ownedFunctions"

    kind = m.EnumPOD("kind", modeltypes.FunctionKind, default="FUNCTION")

    is_leaf = property(lambda self: not self.functions)

    inputs = m.DirectProxyAccessor(FunctionInputPort, aslist=m.ElementList)
    outputs = m.DirectProxyAccessor(FunctionOutputPort, aslist=m.ElementList)

    exchanges: m.Accessor[FunctionalExchange]
    functions: m.Accessor
    packages: m.Accessor
    related_exchanges: m.Accessor[FunctionalExchange]

    realized_functions = m.LinkAccessor[AbstractFunction](
        "ownedFunctionRealizations",
        FunctionRealization,
        aslist=m.MixedElementList,
        attr="targetElement",
    )
    realizing_functions = m.ReferenceSearchingAccessor[AbstractFunction](
        (), "realized_functions", aslist=m.MixedElementList
    )


@m.xtype_handler(None)
class FunctionalExchange(AbstractExchange):
    """A functional exchange."""

    _xmltag = "ownedFunctionalExchanges"

    exchange_items = m.AttrProxyAccessor(
        information.ExchangeItem, "exchangedItems", aslist=m.ElementList
    )

    realized_functional_exchanges = m.LinkAccessor["FunctionalExchange"](
        "ownedFunctionalExchangeRealizations",
        "org.polarsys.capella.core.data.fa:FunctionalExchangeRealization",
        aslist=m.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_functional_exchanges: m.Accessor[FunctionalExchange]

    @property
    def owner(self) -> ComponentExchange | None:
        return self.allocating_component_exchange


class FunctionalChainInvolvement(interaction.AbstractInvolvement):
    """Abstract class for FunctionalChainInvolvementLink/Function."""

    _xmltag = "ownedFunctionalChainInvolvements"


@m.xtype_handler(None)
class FunctionalChainInvolvementLink(FunctionalChainInvolvement):
    """An element linking a FunctionalChain to an Exchange."""

    exchanged_items = m.AttrProxyAccessor(
        information.ExchangeItem, "exchangedItems", aslist=m.ElementList
    )
    exchange_context = m.AttrProxyAccessor(
        capellacore.Constraint, "exchangeContext"
    )


@m.xtype_handler(None)
class FunctionalChainInvolvementFunction(FunctionalChainInvolvement):
    """An element linking a FunctionalChain to a Function."""


@m.xtype_handler(None)
class FunctionalChainReference(FunctionalChainInvolvement):
    """An element linking two related functional chains together."""


@m.xtype_handler(None)
class FunctionalChain(m.GenericElement):
    """A functional chain."""

    _xmltag = "ownedFunctionalChains"

    kind = m.EnumPOD("kind", modeltypes.FunctionalChainKind, default="SIMPLE")

    involvements = m.DirectProxyAccessor(
        m.GenericElement,
        (
            FunctionalChainInvolvementFunction,
            FunctionalChainInvolvementLink,
            FunctionalChainReference,
        ),
        aslist=m.ElementList,
    )
    involved_functions = m.LinkAccessor[AbstractFunction](
        "ownedFunctionalChainInvolvements",
        FunctionalChainInvolvementFunction,
        aslist=m.MixedElementList,
        attr="involved",
    )
    involved_links = m.LinkAccessor[AbstractExchange](
        "ownedFunctionalChainInvolvements",
        FunctionalChainInvolvementLink,
        aslist=m.MixedElementList,
        attr="involved",
    )
    involved_chains = m.LinkAccessor["FunctionalChain"](
        "ownedFunctionalChainInvolvements",
        "org.polarsys.capella.core.data.fa:FunctionalChainReference",
        attr="involved",
        aslist=m.ElementList,
    )
    involving_chains: m.Accessor[FunctionalChain]

    realized_chains = m.LinkAccessor["FunctionalChain"](
        "ownedFunctionalChainRealizations",
        "org.polarsys.capella.core.data.fa:FunctionalChainRealization",
        attr="targetElement",
        backattr="sourceElement",
        aslist=m.ElementList,
    )
    realizing_chains: m.Accessor[FunctionalChain]

    control_nodes = m.DirectProxyAccessor(ControlNode, aslist=m.ElementList)

    @property
    def involved(self) -> m.MixedElementList:
        return self.involved_functions + self.involved_links


@m.xtype_handler(None)
class ComponentPort(m.GenericElement):
    """A component port."""

    _xmltag = "ownedFeatures"

    direction = m.EnumPOD("orientation", modeltypes.OrientationPortKind)
    owner = m.ParentAccessor(m.GenericElement)
    exchanges: m.Accessor


@m.xtype_handler(None)
class ComponentExchange(AbstractExchange):
    """A functional component exchange."""

    _xmltag = "ownedComponentExchanges"

    kind = m.EnumPOD("kind", modeltypes.ComponentExchangeKind, default="UNSET")

    allocated_functional_exchanges = m.LinkAccessor[FunctionalExchange](
        "ownedComponentExchangeFunctionalExchangeAllocations",
        ComponentExchangeFunctionalExchangeAllocation,
        aslist=m.ElementList,
        attr="targetElement",
    )
    allocated_exchange_items = m.AttrProxyAccessor(
        information.ExchangeItem,
        "convoyedInformations",
        aslist=m.ElementList,
    )

    @property
    def owner(self) -> cs.PhysicalLink | None:
        return self.allocating_physical_link

    @property
    def exchange_items(self) -> m.ElementList[information.ExchangeItem]:
        return (
            self.allocated_exchange_items
            + self.allocated_functional_exchanges.map("exchange_items")
        )


for _port, _exchange in [
    (ComponentPort, ComponentExchange),
    (FunctionInputPort, FunctionalExchange),
    (FunctionOutputPort, FunctionalExchange),
]:
    m.set_accessor(
        _port,
        "exchanges",
        m.ReferenceSearchingAccessor(
            _exchange, "source", "target", aslist=m.ElementList
        ),
    )
del _port, _exchange

m.set_accessor(
    ComponentExchange,
    "realized_component_exchanges",
    m.LinkAccessor[ComponentExchange](
        "ownedComponentExchangeRealizations",
        "org.polarsys.capella.core.data.fa:ComponentExchangeRealization",
        aslist=m.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    ),
)
m.set_accessor(
    ComponentExchange,
    "realizing_component_exchanges",
    m.ReferenceSearchingAccessor(
        ComponentExchange, "realized_component_exchanges", aslist=m.ElementList
    ),
)
m.set_accessor(
    FunctionalExchange,
    "allocating_component_exchange",
    m.ReferenceSearchingAccessor(
        ComponentExchange, "allocated_functional_exchanges"
    ),
)
m.set_accessor(
    FunctionalExchange,
    "realizing_functional_exchanges",
    m.ReferenceSearchingAccessor(
        FunctionalExchange,
        "realized_functional_exchanges",
        aslist=m.ElementList,
    ),
)
m.set_accessor(
    FunctionalExchange,
    "involving_functional_chains",
    m.ReferenceSearchingAccessor(
        FunctionalChain, "involved_links", aslist=m.ElementList
    ),
)

m.set_accessor(
    FunctionalChain,
    "involving_chains",
    m.ReferenceSearchingAccessor(
        FunctionalChain, "involved_chains", aslist=m.ElementList
    ),
)
m.set_accessor(
    FunctionalChain,
    "realizing_chains",
    m.ReferenceSearchingAccessor(
        FunctionalChain, "realized_chains", aslist=m.ElementList
    ),
)

m.set_accessor(
    Function,
    "exchanges",
    m.DirectProxyAccessor(FunctionalExchange, aslist=m.ElementList),
)
m.set_accessor(
    Function,
    "related_exchanges",
    m.ReferenceSearchingAccessor(
        FunctionalExchange,
        "source.owner",
        "target.owner",
        aslist=m.ElementList,
    ),
)
m.set_accessor(
    information.ExchangeItem,
    "exchanges",
    m.ReferenceSearchingAccessor(
        (ComponentExchange, FunctionalExchange),
        "exchange_items",
        "allocated_exchange_items",
        aslist=m.ElementList,
    ),
)


from . import cs



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse.model as m

from .. import modeltypes
from . import datavalue


class DataType(m.GenericElement):
    _xmltag = "ownedDataTypes"

    is_discrete = m.BoolPOD("discrete")
    """Whether or not this data type characterizes a discrete value."""
    min_inclusive = m.BoolPOD("minInclusive")
    max_inclusive = m.BoolPOD("maxInclusive")
    pattern = m.StringPOD("pattern")
    """Textual specification of a constraint associated to this data type."""
    visibility = m.EnumPOD(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )


@m.xtype_handler(None)
class BooleanType(DataType):
    literals = m.DirectProxyAccessor(
        datavalue.LiteralBooleanValue,
        aslist=m.ElementList,
        fixed_length=2,
    )
    default = m.RoleTagAccessor("ownedDefaultValue")


@m.xtype_handler(None)
class Enumeration(DataType):
    """An Enumeration."""

    domain_type = m.AttrProxyAccessor(m.GenericElement, "domainType")
    owned_literals = m.DirectProxyAccessor(
        datavalue.EnumerationLiteral, aslist=m.ElementList
    )

    sub: m.Accessor
    super: m.Accessor[Enumeration]

    @property
    def literals(self) -> m.ElementList[datavalue.EnumerationLiteral]:
        """Return all owned and inherited literals."""
        return (
            self.owned_literals + self.super.literals
            if isinstance(self.super, Enumeration)
            else self.owned_literals
        )


@m.xtype_handler(None)
class StringType(DataType):
    default_value = m.RoleTagAccessor("ownedDefaultValue")
    null_value = m.RoleTagAccessor("ownedNullValue")
    min_length = m.RoleTagAccessor("ownedMinLength")
    max_length = m.RoleTagAccessor("ownedMaxLength")


@m.xtype_handler(None)
class NumericType(DataType):
    kind = m.EnumPOD("kind", modeltypes.NumericTypeKind, default="INTEGER")
    default_value = m.RoleTagAccessor("ownedDefaultValue")
    null_value = m.RoleTagAccessor("ownedNullValue")
    min_value = m.RoleTagAccessor("ownedMinValue")
    max_value = m.RoleTagAccessor("ownedMaxValue")


@m.xtype_handler(None)
class PhysicalQuantity(NumericType):
    unit = m.RoleTagAccessor("ownedUnit")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse.model as m


@m.xtype_handler(None)
class Constraint(m.GenericElement):
    """A constraint."""

    _xmltag = "ownedConstraints"

    constrained_elements = m.AttrProxyAccessor(
        m.GenericElement,
        "constrainedElements",
        aslist=m.MixedElementList,
    )

    specification = m.SpecificationAccessor()


@m.xtype_handler(None)
class Generalization(m.GenericElement):
    """A Generalization."""

    _xmltag = "ownedGeneralizations"

    super: m.Accessor


@m.xtype_handler(None)
class EnumerationPropertyLiteral(m.GenericElement):
    """A Literal for EnumerationPropertyType."""

    _xmltag = "ownedLiterals"


@m.xtype_handler(None)
class EnumerationPropertyType(m.GenericElement):
    """An EnumerationPropertyType."""

    _xmltag = "ownedEnumerationPropertyTypes"

    literals = m.DirectProxyAccessor(
        EnumerationPropertyLiteral, aslist=m.ElementList
    )


class PropertyValue(m.GenericElement):
    """Abstract base class for PropertyValues."""

    _xmltag = "ownedPropertyValues"

    enumerations = m.DirectProxyAccessor(
        EnumerationPropertyType, aslist=m.ElementList
    )

    value: m.BasePOD | m.AttrProxyAccessor


@m.xtype_handler(None)
class BooleanPropertyValue(PropertyValue):
    """A boolean property value."""

    value = m.BoolPOD("value")


@m.xtype_handler(None)
class FloatPropertyValue(PropertyValue):
    """A floating point property value."""

    value = m.FloatPOD("value")


@m.xtype_handler(None)
class IntegerPropertyValue(PropertyValue):
    """An integer property value."""

    value = m.IntPOD("value")


@m.xtype_handler(None)
class StringPropertyValue(PropertyValue):
    """A string property value."""

    value = m.StringPOD("value")


@m.xtype_handler(None)
class EnumerationPropertyValue(PropertyValue):
    """An enumeration property value."""

    type = m.AttrProxyAccessor(EnumerationPropertyType, "type")
    value = m.AttrProxyAccessor(EnumerationPropertyLiteral, "value")


@m.xtype_handler(None)
class PropertyValueGroup(m.GenericElement):
    """A group for PropertyValues."""

    _xmltag = "ownedPropertyValueGroups"

    values = m.DirectProxyAccessor(
        m.GenericElement,
        (
            StringPropertyValue,
            BooleanPropertyValue,
            IntegerPropertyValue,
            FloatPropertyValue,
            EnumerationPropertyValue,
        ),
        aslist=m.ElementList,
        mapkey="name",
        mapvalue="value",
    )


@m.xtype_handler(None)
class PropertyValuePkg(m.GenericElement):
    """A Package for PropertyValues."""

    _xmltag = "ownedPropertyValuePkgs"

    enumeration_property_types = m.DirectProxyAccessor(
        EnumerationPropertyType, aslist=m.ElementList
    )
    groups = m.DirectProxyAccessor(
        PropertyValueGroup,
        aslist=m.ElementList,
        mapkey="name",
        mapvalue="values",
    )
    values = m.DirectProxyAccessor(
        m.GenericElement,
        (
            StringPropertyValue,
            BooleanPropertyValue,
            IntegerPropertyValue,
            FloatPropertyValue,
            EnumerationPropertyValue,
        ),
        aslist=m.ElementList,
        mapkey="name",
        mapvalue="value",
    )


m.set_self_references(
    (PropertyValuePkg, "packages"),
)
m.set_accessor(
    m.GenericElement,
    "constraints",
    m.DirectProxyAccessor(Constraint, aslist=m.ElementList),
)

m.set_accessor(
    m.GenericElement,
    "property_values",
    m.DirectProxyAccessor(
        m.GenericElement,
        (
            BooleanPropertyValue,
            EnumerationPropertyValue,
            FloatPropertyValue,
            IntegerPropertyValue,
            StringPropertyValue,
        ),
        aslist=m.MixedElementList,
        mapkey="name",
        mapvalue="value",
    ),
)
m.set_accessor(
    m.GenericElement,
    "property_value_groups",
    m.DirectProxyAccessor(
        PropertyValueGroup,
        aslist=m.ElementList,
        mapkey="name",
        mapvalue="values",
    ),
)

m.set_accessor(
    m.GenericElement,
    "applied_property_values",
    m.AttrProxyAccessor(None, "appliedPropertyValues", aslist=m.ElementList),
)
m.set_accessor(
    m.GenericElement,
    "applied_property_value_groups",
    m.AttrProxyAccessor(
        None, "appliedPropertyValueGroups", aslist=m.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for handling ReqIF Requirements.

.. diagram:: [CDB] Requirements ORM
"""

import typing as t

from ._capellareq import *
from ._glue import *
from ._requirements import *

if not t.TYPE_CHECKING:
    from ._capellareq import __all__ as _cr_all
    from ._requirements import __all__ as _rq_all

    __all__ = [*_cr_all, *_rq_all]
    del _cr_all, _rq_all
del t
